import shutil

import numpy
import pandas
import pandas._testing as tm
import pytest

from tktl.core.managers.project import ProjectManager
from tktl.core.t import TemplateT


@pytest.fixture
def serializer_df_inputs():
    # from: pandas.tests.frame.conftest.float_frame_with_na
    df = pandas.DataFrame(tm.getSeriesData())
    # set some NAs
    df.iloc[5:10] = numpy.nan
    df.iloc[15:20, -2:] = numpy.nan
    return df


@pytest.fixture
def serializer_series_inputs(serializer_df_inputs):
    return serializer_df_inputs.B


@pytest.fixture
def serializer_l_of_d_sequence_inputs():
    return [
        {"a": i, "b": j}
        for i, j in zip(range(100), numpy.random.randn(100).tolist())  # noqa
    ]


@pytest.fixture
def serializer_dict_inputs():
    return {"a": 1000, "b": 1.0, "c": "strfgadfsv"}


@pytest.fixture
def serializer_list_and_tuple_inputs():
    return (
        numpy.random.randn(100).tolist(),
        [numpy.random.randn(100).tolist(), numpy.random.randn(100).tolist()],
        [1, 2, 2],
        ["fdefswdefswde", "dfasfwefgwe", "fjnlsjnk2"],
        (1, 2, 2, 1, 2323, 123, 12, 342, 14, 1234, 123, 43),
    )


@pytest.fixture
def serializer_list_inputs():
    return numpy.random.randn(100).tolist()


@pytest.fixture(scope="function")
def create_proj():
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    yield
    shutil.rmtree("sample_project")



import pathlib
import typing as t

import pandas as pd  # type: ignore
import pyarrow.parquet as pq  # type: ignore

from tktl.core.loggers import LOG

JSONObject = t.Dict[t.AnyStr, t.Any]
JSONArray = t.List[t.Any]
JSONStructure = t.Union[JSONArray, JSONObject]


def coerce_dataframe_to_series(df: pd.DataFrame) -> t.Union[pd.DataFrame, pd.Series]:
    if len(df.columns) == 1:
        # Since we store dataframes in parquet files, we assume that a
        # one column dataframe used to be a Series.
        return df.iloc[:, 0]
    return df


def coerce_series_to_dataframe(obj: t.Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
    if isinstance(obj, pd.Series):
        return pd.DataFrame({str(obj.name): obj})
    if len(obj.columns) == 1:
        LOG.error(
            "Storing 1 column dataframe, this will become a Series on the way back"
        )
    return obj


def get_number_of_rows_in_chunk(path: pathlib.Path, bytes_per_chunk: int) -> int:
    """get_chunk_size.
    Calculate number of rows per chunk. This file does _not_ read the
    data from the parquet file.

    If the file is 1024 bytes and every chunk is 500 bytes we want to have
    3 chunks. Let's say that there are 16 rows in the file, then the desired
    result is 6 rows per chunk.

    Parameters
    ----------
    path : pathlib.Path
        path to the parquet file
    bytes_per_chunk : int
        bytes that are supposed to be in every chunk

    Returns
    -------
    int - number of rows every chunk contains

    """
    file_size_in_bytes = path.stat().st_size
    pq_file = pq.ParquetFile(path)
    number_of_rows = pq_file.metadata.num_rows

    chunks = file_size_in_bytes // bytes_per_chunk + 1

    return number_of_rows // chunks + 1



import requests

from tktl import __version__ as tktl_version
from tktl.core import utils
from tktl.core.loggers import LOG


class API:
    DEFAULT_HEADERS = {
        "Accept": "application/json",
        "User-Agent": f"taktile-cli:{tktl_version}",
        "Content-Type": "application/json",
    }

    def __init__(self, api_url, headers={}):
        """

        Parameters
        ----------
        api_url
        headers
        """

        self._api_url = api_url
        self._headers = {**self.DEFAULT_HEADERS, **headers}

    def get_path(self, url=None):
        if not url:
            return self._api_url
        full_path = utils.concatenate_urls(self._api_url, url)
        return full_path

    def post(
        self, url=None, json=None, params=None, files=None, data=None, timeout=None
    ):
        path = self.get_path(url)

        LOG.trace(
            "POST request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}\n\tfiles: {}\n\tdata: {}".format(
                path, self._headers, json, params, files, data
            )
        )
        response = requests.post(
            path,
            json=json,
            params=params,
            headers=self._headers,
            files=files,
            data=data,
            timeout=timeout,
        )
        LOG.trace("Response status code: {}".format(response.status_code))
        LOG.trace("Response content: {}".format(response.content))
        return response

    def put(self, url, json=None, params=None, data=None):
        path = self.get_path(url)
        LOG.trace(
            "PUT request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self._headers, json, params
            )
        )
        response = requests.put(
            path, json=json, params=params, headers=self._headers, data=data
        )
        LOG.trace("Response status code: {}".format(response.status_code))
        LOG.trace("Response content: {}".format(response.content))
        return response

    def patch(self, url, json=None, params=None, data=None):
        path = self.get_path(url)
        LOG.trace(
            "PATCH request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self._headers, json, params
            )
        )
        response = requests.patch(
            path, json=json, params=params, headers=self._headers, data=data
        )
        LOG.trace("Response status code: {}".format(response.status_code))
        LOG.trace("Response content: {}".format(response.content))
        return response

    def get(self, url, json=None, params=None):
        path = self.get_path(url)
        LOG.trace(
            "GET request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self._headers, json, params
            )
        )
        response = requests.get(path, params=params, headers=self._headers, json=json)
        LOG.trace("Response status code: {}".format(response.status_code))
        LOG.trace("Response content: {}".format(response.content))
        return response

    def delete(self, url, json=None, params=None):
        path = self.get_path(url)
        response = requests.delete(
            path, params=params, headers=self._headers, json=json
        )
        LOG.trace(
            "DELETE request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                response.url, self._headers, json, params
            )
        )
        LOG.trace("Response status code: {}".format(response.status_code))
        LOG.trace("Response content: {}".format(response.content))
        return response



import logging
import os
from typing import cast

from tktl.core.exceptions import validate_config
from tktl.core.managers.base import BaseConfigManager
from tktl.core.managers.project import ProjectManager

log = logging.getLogger("root")


class LearnerManager(BaseConfigManager):
    CONFIG_FILE_NAME: str = "tktl.yaml"
    INPUT_KEYS: str = "x"
    OUTPUT_KEYS: str = "y"
    PLATFORM: str = "tensorflow"

    @classmethod
    @validate_config
    def get_path(cls, name: str) -> str:
        return os.path.join(cast(str, ProjectManager.get_value("saved-models")), name)



from .abc import Endpoint, EndpointResponse, TaskType  # noqa: F401
from .arrow import ArrowEndpoint  # noqa: F401
from .generic import GenericEndpoint  # noqa: F401
from .profiled import ProfiledEndpoint  # noqa: F401
from .typed import TypedEndpoint  # noqa: F401



import json

import numpy as np  # type: ignore

from tktl.future.utils import JSONStructure


def numpy_deserialize(*, value: JSONStructure, sample: np.ndarray) -> np.ndarray:
    return np.array(value)


def numpy_serialize(*, value: np.ndarray) -> str:
    return json.dumps(value.tolist())


def numpy_to_example(*, value: np.ndarray) -> str:
    return json.dumps(value[0:1].tolist())



import os
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal, assert_series_equal
from taktile_types.enums.endpoint import EndpointKinds

from tktl.core.clients.future.endpoint import SDKRestEndpoint
from tktl.core.clients.http_client import API
from tktl.core.clients.rest import RestClient
from tktl.core.exceptions import TaktileSdkError
from tktl.core.future.t import ArrowFormatKinds
from tktl.core.managers.auth import AuthConfigManager
from tktl.core.schemas.future.service import EndpointInfoSchema


def _assert_equal(a, b):
    assert a == b


RETURN_VALUE = [0.88]


def test_instantiate_client():

    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)

    with pytest.raises(TaktileSdkError):
        RestClient(
            api_key=key,
            repository_name=f"{os.environ['TEST_USER']}/test-new",
            branch_name="master",
            endpoint_name="repayment",
        )

    client = RestClient(
        api_key=key,
        repository_name=f"{os.environ['TEST_USER']}/integ-testing",
        branch_name="master",
        endpoint_name="repayment",
    )
    assert client.location is not None


def test_instantiate_by_url(sample_deployed_url):
    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)
    client = RestClient.for_url(
        api_key=key, url=sample_deployed_url, endpoint_name="repayment"
    )
    assert client.location == f"https://{sample_deployed_url}/"


@pytest.mark.parametrize(
    "X,X_serialized",
    [(pd.DataFrame([1]), '[{"0":1}]'), (pd.Series([1]), "[1]"), (np.array([1]), "[1]")],
)
@pytest.mark.parametrize(
    "response_kind,result,assert_equal",
    [
        (None, RETURN_VALUE, _assert_equal),
        (ArrowFormatKinds.DATAFRAME, pd.DataFrame(RETURN_VALUE), assert_frame_equal),
        (ArrowFormatKinds.SERIES, pd.Series(RETURN_VALUE), assert_series_equal),
        (ArrowFormatKinds.ARRAY, np.array(RETURN_VALUE), assert_array_equal),
    ],
)
def test_rest_client_serialization(
    X, X_serialized, response_kind, result, assert_equal
):
    info: EndpointInfoSchema = EndpointInfoSchema(
        name="dummy",
        path="dummy",
        kind=EndpointKinds.ARROW,
        response_kind=response_kind,
    )
    api = API(api_url="dummy")
    rest_endpoint = SDKRestEndpoint(api=api, info=info)
    with patch("tktl.core.clients.future.endpoint.safe_post") as mock:
        mock.return_value = RETURN_VALUE
        return_val = rest_endpoint.__call__(X)
        mock.assert_called_with(
            api=api, url=info.path, retries=3, timeout=10.0, data=X_serialized
        )
        assert_equal(return_val, result)


RETURN_VALUE_DT = ["2015-12-31T00:00:00.000Z"]


@pytest.mark.parametrize(
    "response_kind,result,assert_equal",
    [
        (None, RETURN_VALUE_DT, _assert_equal),
        (ArrowFormatKinds.DATAFRAME, pd.DataFrame(RETURN_VALUE_DT), assert_frame_equal),
        (ArrowFormatKinds.SERIES, pd.Series(RETURN_VALUE_DT), assert_series_equal),
        (ArrowFormatKinds.ARRAY, np.array(RETURN_VALUE_DT), assert_array_equal),
    ],
)
def test_rest_client_datetime(response_kind, result, assert_equal):
    X = pd.DataFrame([pd.to_datetime(["2015-12-31"])], columns=["end_date"])
    X_ser = '[{"end_date":"2015-12-31T00:00:00.000Z"}]'
    info: EndpointInfoSchema = EndpointInfoSchema(
        name="dummy",
        path="dummy",
        kind=EndpointKinds.ARROW,
        response_kind=response_kind,
    )
    api = API(api_url="dummy")
    rest_endpoint = SDKRestEndpoint(api=api, info=info)
    with patch("tktl.core.clients.future.endpoint.safe_post") as mock:
        mock.return_value = RETURN_VALUE_DT
        return_val = rest_endpoint.__call__(X)
        mock.assert_called_with(
            api=api, url=info.path, retries=3, timeout=10.0, data=X_ser
        )
        assert_equal(return_val, result)



import requests
from pyarrow._flight import FlightCancelledError  # type: ignore
from pyarrow.flight import FlightClient  # type: ignore

from tktl.commands import BaseGrpcClientCommand, BaseRestClientCommand
from tktl.core.clients import API
from tktl.core.exceptions import APIClientException


class GetRestHealthCommand(BaseRestClientCommand):
    def execute(self):
        if self.client.local:
            self.client.set_client_and_location(
                location=self.client.local_endpoint,
                client=API(api_url=self.client.local_endpoint),
            )
        else:
            self.client._authenticate(health_check=True)
        try:
            response = self.client.health()
        except APIClientException as e:
            error_str = f"Service is not running properly: {repr(e.detail)}"
            self.client.logger.error(error_str)
            raise APIClientException(detail=error_str, status_code=e.status_code)
        except requests.exceptions.RequestException as e:
            error_str = f"Unable to decode server response: {repr(e)}"
            self.client.logger.error(error_str)
            raise APIClientException(detail=error_str, status_code=400)
        if response.status_code < 400:
            self.client.logger.log(
                f"Service at {self.client.location} is up and running", color="green"
            )
            return response


class GetGrpcHealthCommand(BaseGrpcClientCommand):
    def execute(self):
        if self.client.local:
            self.client.set_client_and_location(
                location=self.client.local_endpoint,
                client=FlightClient(location=self.client.local_endpoint),
            )
        else:
            try:
                self.client._authenticate(health_check=True)
            except FlightCancelledError as e:
                error_str = f"Service is not running properly: {repr(e)}"
                self.client.logger.error(error_str)
                raise APIClientException(detail=error_str, status_code=404)
        try:
            response = self.client.health()
        except APIClientException as e:
            error_str = f"Service is not running properly: {repr(e.detail)}"
            self.client.logger.error(error_str)

            raise APIClientException(detail=error_str, status_code=e.status_code)

        self.client.logger.log(
            f"Service at {self.client.location} is up and running", color="green"
        )
        return response



import os
import shutil

import pytest

from tktl.commands.init import init_project
from tktl.commands.validate import validate_project_config
from tktl.core.managers.constants import TEMPLATE_PROJECT_DIRS, TEMPLATE_PROJECT_FILES
from tktl.core.managers.project import ProjectManager
from tktl.core.t import TemplateT


def test_create_in_existing_directory(monkeypatch, tmpdir):
    monkeypatch.setattr("builtins.input", lambda: "Y")
    ProjectManager.init_project(
        tmpdir.dirpath(), "test_project_dir", TemplateT.REPAYMENT
    )
    for d in TEMPLATE_PROJECT_DIRS:
        assert os.path.isdir(os.path.join(tmpdir.dirpath(), "test_project_dir", d))
    for file in TEMPLATE_PROJECT_FILES[TemplateT.REPAYMENT]:
        assert os.path.exists(os.path.join(tmpdir.dirpath(), "test_project_dir", file))
    assert os.path.exists(
        os.path.join(tmpdir.dirpath(), "test_project_dir", "tktl.yaml")
    )


def test_create_project():
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    for file in TEMPLATE_PROJECT_FILES[TemplateT.REPAYMENT]:
        assert os.path.exists(os.path.join("sample_project", file))
    assert os.path.exists(os.path.join("sample_project", "tktl.yaml"))

    for d in TEMPLATE_PROJECT_DIRS:
        assert os.path.isdir(os.path.join("sample_project", d))
    shutil.rmtree("sample_project")


def test_safe_init(monkeypatch, caplog):
    monkeypatch.setattr("builtins.input", lambda: "n")

    with pytest.raises(SystemExit):
        ProjectManager.safe_init(None, "sample_project", TemplateT.REPAYMENT)
        assert "Path specified already exists" in caplog.text


def test_overwrite(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", lambda: "Y")
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    init_project(None, "sample_project", TemplateT.REPAYMENT)
    out, err = capsys.readouterr()
    assert "Project scaffolding created successfully" in out
    shutil.rmtree("sample_project")


def test_validate(capsys):
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert "Project scaffolding is valid!\n" in out
    shutil.rmtree("sample_project")

    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    shutil.rmtree("sample_project/src")
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert err is not None
    assert "Project scaffolding is invalid: Missing Files or Directory ❌\n" in out
    assert "Missing directories in repository: src" in err
    shutil.rmtree("sample_project")

    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    with open("sample_project/tktl.yaml", "w") as inf:
        inf.write(
            """
default_branch_name: master
service:
  replicas: 3
  compute_type: CPUqqqqqqqq
  endpoint_type: RESTuuuuuu
"""
        )
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert "Project scaffolding is invalid: Invalid Config File ❌\n" in out
    shutil.rmtree("sample_project")



import typing as t

from pyarrow.flight import ClientAuthHandler, FlightClient  # type: ignore

from tktl.core.clients.future.taktile import DeploymentApiClient
from tktl.core.config import settings
from tktl.core.loggers import LOG
from tktl.core.schemas.repository import (
    RepositoryDeployment,
    _format_grpc_url,
    load_certs,
)

from .endpoint import ArrowEndpoints


class ApiKeyClientAuthHandler(ClientAuthHandler):
    def __init__(self, api_key: str):
        super().__init__()
        self._api_key = api_key

    def authenticate(self, outgoing, incoming):
        outgoing.write(self._api_key)
        self._api_key = incoming.read()

    def get_token(self):
        return self._api_key


class ArrowClient:

    endpoints: ArrowEndpoints

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str,
        url: t.Optional[str] = None,
    ):
        self._api_key = api_key
        self._repository_name = repository_name
        self._branch_name = branch_name
        self._url = ArrowClient.__get_url(
            url=url,
            repository_name=repository_name,
            branch_name=branch_name,
            api_key=api_key,
        )

        self._client = ArrowClient.__get_deployment_client(
            url=self._url, api_key=self._api_key
        )

        self._actions = self._client.list_actions()

        self.endpoints = ArrowEndpoints(client=self._client, actions=self._actions)

    @staticmethod
    def __get_deployment(
        *, repository_name: str, branch_name: str, client: DeploymentApiClient
    ) -> RepositoryDeployment:
        return client.get_deployment_by_branch_name(
            repository_name=repository_name, branch_name=branch_name
        )

    @staticmethod
    def __get_dapi_client(*, api_key: str) -> DeploymentApiClient:
        client = DeploymentApiClient(api_key)

        return client

    @staticmethod
    def __get_deployment_client(*, url: str, api_key: str) -> FlightClient:
        certs = None if settings.LOCAL_STACK else load_certs()
        client = FlightClient(location=url, tls_root_certs=certs)
        LOG.trace(f"Performing authentication request against {url}")
        client.authenticate(ApiKeyClientAuthHandler(api_key=api_key))

        return client

    @staticmethod
    def __get_url(
        *, url: t.Optional[str], repository_name: str, branch_name: str, api_key: str
    ) -> str:
        if url:
            return url

        client = ArrowClient.__get_dapi_client(api_key=api_key)
        deployment = ArrowClient.__get_deployment(
            repository_name=repository_name, branch_name=branch_name, client=client
        )

        return _format_grpc_url(deployment.public_docs_url)



import typing as t

from taktile_types.enums.endpoint import EndpointKinds

from .abc import Endpoint, XType, YType


class GenericEndpoint(Endpoint):
    kind: EndpointKinds = EndpointKinds.GENERIC

    @staticmethod
    def supported(
        *, X: XType = None, y: YType = None, profile: t.Optional[str] = None,
    ) -> bool:
        return profile is None and X is None and y is None



from collections.abc import Sequence

import pandas  # type: ignore


def get_list_shape(sequence, inner_types=(), shape=()):
    """
    returns the shape of nested lists similarly to numpy's shape.
    """
    if not isinstance(sequence, Sequence):
        return shape, inner_types

    if isinstance(sequence[0], Sequence):
        length = len(sequence[0])
        if not all(len(item) == length for item in sequence):
            raise ValueError("Lists with inconsistent shapes")
        types = ()
    else:

        types = ((next(type(v) for v in sequence if not pandas.isna(v))),)
    shape += (len(sequence),)
    shape, types = get_list_shape(sequence[0], inner_types=types, shape=shape)
    return shape, types


def is_valid_list_of_lists(lst):
    is_list_of_lists = all([isinstance(i, list) for i in lst])
    if not is_list_of_lists:
        return False
    lengths = [len(i) for i in lst]
    return lengths.count(lengths[0]) == len(lengths)



from typing import Optional

from tktl.core.loggers import LOG
from tktl.core.managers.project import ProjectManager
from tktl.core.t import TemplateT


def init_project(path: Optional[str], name: str, template: TemplateT):
    project_path = ProjectManager.init_project(path, name, template)
    LOG.log(
        f"Project scaffolding created successfully at {project_path}", color="green",
    )



import click

from tktl.commands import login as login_commands


@click.command("version", help="Show the version and exit")
@click.option(
    "--check/--no-check", default=True, help="Check if new version is available"
)
def get_version(check: bool = True):
    command = login_commands.ShowVersionCommand()
    command.execute(check=check)



import pandas as pd  # type: ignore

from tktl.future.utils import JSONStructure


def series_deserialize(*, value: JSONStructure, sample: pd.Series) -> pd.Series:
    return pd.Series(value)


def series_serialize(*, value: pd.Series) -> str:
    return value.to_json(orient="records", date_format="iso")


def series_to_example(*, value: pd.Series) -> str:
    return value.iloc[[0]].to_json(orient="records", date_format="iso")



import click

from tktl.cli.common import (
    ClickGroup,
    get_secrets,
    validate_integration_shared_options,
    validate_project_shared_options,
)
from tktl.commands.validate import (
    build_image,
    get_docker_manager,
    validate_endpoint_sample,
    validate_import,
    validate_integration,
    validate_profiling,
    validate_project_config,
    validate_unittest,
)
from tktl.core.config import settings


@click.group(
    "validate", help="Validate the project", cls=ClickGroup, **settings.HELP_COLORS_DICT
)
def validate():
    pass


@validate.command("config", help="Validate the configuration")
@click.option(
    "--path", "-p", help="Validate project located at this path", type=str, default="."
)
def validate_config_command(path) -> None:
    """Validates a new project for the necessary scaffolding, as well as the supporting
    files needed. The directory structure of a new project.
    """
    validate_project_config(path=path)


@validate.command("import", help="Validate src/endpoints.py")
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_import_command(
    path: str, cache: bool, prune: bool, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)
    validate_import(dm=dm, image=image)
    if prune:
        dm.remove_image(image)


@validate.command("sample", help="Validate Sample Data on Endpoints")
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_sample_command(
    path: str, cache: bool, prune: bool, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)
    validate_endpoint_sample(dm=dm, image=image)
    if prune:
        dm.remove_image(image)


@validate.command("unittest", help="Validate the unittests")
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_unittest_command(
    path: str, cache: bool, prune: bool, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)
    validate_unittest(dm=dm, image=image)
    if prune:
        dm.remove_image(image)


@validate.command("integration", help="Validate integration")
@validate_integration_shared_options
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_integration_command(
    path: str,
    cache: bool,
    prune: bool,
    timeout: int,
    retries: int,
    secrets_repository: str,
) -> None:
    secrets = get_secrets(secrets_repository)
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)
    validate_integration(dm=dm, image=image, timeout=timeout, retries=retries)
    if prune:
        dm.remove_image(image)


@validate.command("profiling", help="Validate profiling")
@validate_integration_shared_options
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_profiling_command(
    path: str,
    cache: bool,
    prune: bool,
    timeout: int,
    retries: int,
    secrets_repository: str,
) -> None:
    secrets = get_secrets(secrets_repository)
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)
    validate_profiling(
        dm=dm, image=image, timeout=timeout, retries=retries, prune=prune
    )
    if prune:
        dm.remove_image(image)


@validate.command("all", help="Validate everything")
@validate_integration_shared_options
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_all_command(
    path: str,
    cache: bool,
    prune: bool,
    timeout: int,
    retries: int,
    secrets_repository: str,
) -> None:
    validate_project_config(path=path)

    secrets = get_secrets(secrets_repository)
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)

    validate_import(dm=dm, image=image)
    validate_endpoint_sample(dm=dm, image=image)
    validate_unittest(dm=dm, image=image)
    validate_integration(dm=dm, image=image, timeout=timeout, retries=retries)
    validate_profiling(
        dm=dm, image=image, timeout=timeout, retries=retries, prune=prune
    )

    if prune:
        dm.remove_image(image)



import pathlib
import typing as t

import pandas as pd  # type: ignore
import pyarrow as pa  # type: ignore
import pyarrow.parquet as pq  # type: ignore

from tktl.core.exceptions import UnlazifiedParquetError
from tktl.core.loggers import LOG
from tktl.core.t import LazifyType

TAKTILE_LAZYLOAD_VERSION_KEY = "taktile_version".encode()
TAKTILE_LAZYLOAD_VERSION_VALUE = "0.2.0".encode()
TAKTILE_LAZYLOAD_BACKGROUND_DATA_KEY = "taktile_background_data".encode()
TAKTILE_LAZYLOAD_TYPE_KEY = "taktile_lazyload_type_key".encode()


def load_background_data_from_lazified_pq(path: pathlib.Path) -> pd.DataFrame:
    schema = pq.read_schema(path)

    if TAKTILE_LAZYLOAD_VERSION_KEY not in schema.metadata:
        raise UnlazifiedParquetError(f"{path} is not a lazified parquet file")

    background_data_bytes = schema.metadata[TAKTILE_LAZYLOAD_BACKGROUND_DATA_KEY]
    background_data = pd.read_json(background_data_bytes, orient="table")

    return background_data


def lazify_parquet_metadata(
    table: pa.Table,
    data: t.Optional[pd.DataFrame] = None,
    lazify_type: LazifyType = LazifyType.SHAP,
) -> pa.Table:
    """lazify_parquet_metadata.
    Add lazy loading information to parquet schema. If
    parameter data is provided, this data is stored. Otherwise
    background data is calculated and stored.

    Parameters
    ----------
    table : pa.Table
        table to lazify
    data : t.Optional[pd.DataFrame]
        data is the lazification data if provided
    lazify_type: LazifyType
        the type of lazification to be performed

    Returns
    -------
    pa.Table the same table with patched metadata

    """

    if data is None:
        try:
            from profiling.shap import ShapExplainer  # type: ignore
        except ImportError:
            LOG.error(
                "Could not load profiling package. Please make sure profiling is installed to lazify metadata"
            )

        df = table.to_pandas()

        LOG.trace("Calculating background data...")
        if lazify_type == LazifyType.SHAP:
            background_data = ShapExplainer._create_background_data_from_full_frame(df)
        else:
            background_data = _create_background_data_as_first_valid_occurence(df)
        background_data_json = background_data.to_json(orient="table")
    else:
        background_data_json = data.to_json(orient="table")

    if TAKTILE_LAZYLOAD_VERSION_KEY in table.schema.metadata:
        LOG.warning("Metadata is already lazified")

    LOG.trace("Patching metadata ...")
    metadata = {
        **table.schema.metadata,
        TAKTILE_LAZYLOAD_VERSION_KEY: TAKTILE_LAZYLOAD_VERSION_VALUE,
        TAKTILE_LAZYLOAD_BACKGROUND_DATA_KEY: background_data_json,
        TAKTILE_LAZYLOAD_TYPE_KEY: lazify_type.value,
    }

    return table.cast(table.schema.with_metadata(metadata))


def _create_background_data_as_first_valid_occurence(df):
    row = {}
    for col in df.columns:
        val = df[col].get(df[col].first_valid_index())
        row[col] = (
            val.dropna().iloc[0] if isinstance(val, pd.Series) else val
        )  # Duplicate Index
    return pd.DataFrame([row], columns=df.columns).astype(df.dtypes)



import math
from typing import Any, Optional, Tuple

import pandas  # type: ignore
from pyarrow import Table  # type: ignore
from pyarrow.flight import ClientAuthHandler  # type: ignore
from pyarrow.flight import (
    FlightCancelledError,
    FlightClient,
    FlightDescriptor,
    FlightInfo,
    FlightUnauthenticatedError,
    FlightUnavailableError,
    Ticket,
)
from pyarrow.lib import ArrowException  # type: ignore
from taktile_types.enums.instance import ServiceType

from tktl.core.clients import Client
from tktl.core.config import settings
from tktl.core.exceptions import APIClientException, AuthenticationError
from tktl.core.loggers import LOG, Logger
from tktl.core.schemas.repository import _format_grpc_url, load_certs
from tktl.core.serializers import deserialize_arrow, serialize_arrow


class ApiKeyClientAuthHandler(ClientAuthHandler):
    """An example implementation of authentication via ApiKey."""

    def __init__(self, api_key: str):
        super(ApiKeyClientAuthHandler, self).__init__()
        self.api_key = api_key

    def authenticate(self, outgoing, incoming):
        outgoing.write(self.api_key)
        self.api_key = incoming.read()

    def get_token(self):
        return self.api_key


class ArrowFlightClient(Client):
    TRANSPORT = ServiceType.GRPC

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str,
        endpoint_name: str,
        local: bool = False,
        logger: Logger = LOG,
        verbosity: int = 0,
        skip_auth: bool = False,
    ):
        super().__init__(
            api_key=api_key,
            repository_name=repository_name,
            branch_name=branch_name,
            endpoint_name=endpoint_name,
            local=local,
            logger=logger,
            verbosity=verbosity,
        )
        if not skip_auth:
            try:
                self._authenticate()
            except ArrowException as AE:
                self.logger.error(
                    f"Unable to authenticate against Arrow endpoint: {AE}"
                )
                exit(1)

    @classmethod
    def for_url(cls, api_key: str, url: str, endpoint_name: str):
        grpc_url = cls.format_url(url=url)
        instantiated = cls(
            api_key=api_key,
            repository_name="",
            branch_name="",
            endpoint_name=endpoint_name,
            skip_auth=True,
        )
        if settings.LOCAL_STACK:
            certs = None
        else:
            certs = load_certs()
        client = FlightClient(tls_root_certs=certs, location=grpc_url)
        instantiated.logger.trace(
            f"Performing authentication request against {grpc_url}"
        )
        client.authenticate(
            ApiKeyClientAuthHandler(api_key=instantiated.taktile_client.api_key)
        )
        instantiated.set_client_and_location(location=grpc_url, client=client)
        return instantiated

    @staticmethod
    def format_url(url: Optional[str]) -> str:
        return _format_grpc_url(url)

    @property
    def local_endpoint(self):
        return settings.LOCAL_ARROW_ENDPOINT

    def list_deployments(self):
        pass

    def list_commands(self):
        return self.client.list_actions()

    def _authenticate(self, health_check: bool = False):
        if health_check:
            location = self.get_deployment_location()
        else:
            try:
                location = self.get_endpoint_and_location()
            except APIClientException as e:
                self.logger.error(f"Unable to authenticate: {e.detail}")
                exit(1)
                return

        certs = None if settings.LOCAL_STACK else load_certs()
        client = FlightClient(location=location, tls_root_certs=certs)
        self.logger.trace(f"Performing authentication request against {location}")
        client.authenticate(
            ApiKeyClientAuthHandler(api_key=self.taktile_client.api_key)
        )

        self.set_client_and_location(location=location, client=client)

    def _send_batch(self, writer, batch, reader, batch_number):
        try:
            writer.write_batch(batch)
            return reader.read_chunk()
        except Exception as e:
            self.logger.error(
                f"ERROR: performing prediction for batch {batch_number}: {e} "
                f"The predictions from this batch will be missing from the result"
            )
            return None

    def predict(self, inputs: Any, use_input_index: bool = False) -> Any:
        table = serialize_arrow(inputs)
        batch_size, batch_memory = get_chunk_size(table)
        if not (batch_size and batch_memory):
            return
        descriptor = self.get_flight_info(command_name=str.encode(self.endpoint_name))
        writer, reader = self.client.do_exchange(descriptor.descriptor)  # type: ignore
        self.logger.trace(
            f"Initiating prediction request with batches of {batch_size} records of "
            f"~{batch_memory:.2f} MB/batch"
        )
        batches = table.to_batches(max_chunksize=batch_size)
        chunks = []
        schema = None
        with writer:
            writer.begin(table.schema)
            for i, batch in enumerate(batches):
                self.logger.trace(f"Prediction for batch {i + 1}/{len(batches)}")
                chunk = self._send_batch(
                    writer=writer, batch=batch, reader=reader, batch_number=i + 1
                )
                if not chunk:
                    continue
                if not schema and chunk.data.schema is not None:
                    schema = chunk.data.schema
                chunks.append(chunk.data)
        deserialized = deserialize_arrow(Table.from_batches(chunks, schema))
        if use_input_index:
            input_has_index = isinstance(inputs, pandas.Series) or isinstance(
                inputs, pandas.DataFrame
            )
            output_has_index = isinstance(deserialized, pandas.Series) or isinstance(
                deserialized, pandas.DataFrame
            )
            if not input_has_index or not output_has_index:
                LOG.warning(
                    "Inputs or Outputs are not of type series or dataframe, use_input_index has no effect"
                )
            else:
                try:
                    deserialized.index = inputs.index
                except Exception as e:
                    LOG.warning(f"Unable to set indexes of output frame: {repr(e)}")
        return deserialized

    def get_sample_data(self):
        if not self.endpoint_name:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        if not self.endpoint_has_arrow_sample_data:
            self.logger.warning("Sample data not available for this endpoint")
            return None, None
        schema = self.get_schema()
        if schema.metadata is None:
            self.logger.error(
                "Sample data does not exist or is not supported for this endpoint"
            )
            return None, None
        x_ticket, y_ticket = (
            str.encode(f"{self.endpoint_name}__X"),
            str.encode(f"{self.endpoint_name}__y"),
        )
        return self._get_data(ticket=x_ticket), self._get_data(ticket=y_ticket)

    def _get_data(self, ticket):
        if not self.endpoint_name:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        self.logger.trace("Fetching sample data from server")
        reader = self.client.do_get(Ticket(ticket=ticket))
        return deserialize_arrow(reader.read_all())

    def get_schema(self):
        if not self.endpoint_name:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        info = self.get_flight_info(str.encode(self.endpoint_name))
        return info.schema

    def get_flight_info(self, command_name: bytes) -> FlightInfo:
        descriptor = FlightDescriptor.for_command(command_name)
        return self.client.get_flight_info(descriptor)  # type: ignore

    def health(self):
        try:
            self.logger.trace("Connecting to server...")
            self.client.wait_for_available(timeout=1)
        except FlightUnauthenticatedError:
            self.logger.trace("Connection successful")
            return True
        except (FlightCancelledError, FlightUnavailableError):
            raise APIClientException(
                detail="Arrow flight is unavailable", status_code=502
            )
        return True


def get_chunk_size(sample_table: Table) -> Tuple[Optional[int], Optional[float]]:
    try:
        mem_per_record = sample_table.nbytes / sample_table.num_rows
    except ZeroDivisionError:
        LOG.error(
            "Empty payload received, which is currently not supported for arrow endpoints"
        )
        return None, None
    batch_size = math.ceil(settings.ARROW_BATCH_MB * 1e6 / mem_per_record)
    batch_memory_mb = (batch_size * mem_per_record) / 1e6
    return batch_size, batch_memory_mb



import typing as t

from pydantic import BaseModel
from taktile_types.enums.endpoint import EndpointKinds

from .abc import XType, YType
from .generic import GenericEndpoint


class TypedEndpoint(GenericEndpoint):
    kind: EndpointKinds = EndpointKinds.TYPED

    @staticmethod
    def supported(
        *, X: XType = None, y: YType = None, profile: t.Optional[str] = None,
    ) -> bool:
        return (
            profile is None
            and isinstance(X, (type(BaseModel), type(t.Any)))
            and isinstance(y, (type(BaseModel), type(t.Any)))
        )



import functools
from typing import Dict, Optional

import click
from click_didyoumean import DYMMixin  # type: ignore
from click_help_colors import HelpColorsCommand  # type: ignore
from click_help_colors import HelpColorsGroup
from pydantic import UUID4

from tktl.core.loggers import LOG  # noqa: F401
from tktl.core.managers.secrets import SecretsManager

OPTIONS_FILE_OPTION_NAME = "optionsFile"
OPTIONS_FILE_PARAMETER_NAME = "options_file"
OPTIONS_DUMP_FILE_OPTION_NAME = "createOptionsFile"


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))


class ClickGroup(DYMMixin, HelpColorsGroup, AliasedGroup):
    pass


class ClickCommand(DYMMixin, HelpColorsCommand):
    pass


def deprecated(msg):
    deprecated_invoke_notice = (
        msg
        + """\nFor more information, please see:

https://docs.taktile.com
If you depend on functionality not listed there, please file an issue."""
    )

    def new_invoke(self, ctx):
        click.echo(click.style(deprecated_invoke_notice, fg="red"), err=True)
        super(type(self), self).invoke(ctx)

    def decorator(f):
        f.invoke = functools.partial(new_invoke, f)

    return decorator


_shared_options = [
    click.option("-f", "--full", help="Full (verbose) output", is_flag=True),
    click.option(
        "-a", "--all", "all_resources", help="Return all resources", is_flag=True
    ),
    click.option(
        "-O",
        "--output",
        help="Output as json or yaml ot stdout",
        default="stdout",
        type=click.Choice(["stdout", "json", "yaml"]),
    ),
]


_filter_deploy_shared_options = [
    click.option("-c", "--commit", help="Commit hash"),
    click.option("-b", "--branch", help="Branch name"),
    click.option("-s", "--status", help="Status in which deployment is"),
]


_filter_repo_shared_options = [
    click.option("-r", "--repo", help="Repository name"),
    click.option("-o", "--owner", help="Owner of the repository"),
]

_validate_project_shared_options = [
    click.option(
        "--path",
        "-p",
        help="Validate project located at this path",
        type=str,
        default=".",
    ),
    click.option(
        "--cache/--disable-cache",
        help="Enable or disable use of Docker cache",
        default=True,
    ),
    click.option(
        "--prune/--disable-prune",
        envvar="PRUNE_IMAGES",
        help="Enable or disable automatic pruning images",
        default=False,
    ),
]

_validate_integration_shared_options = [
    click.option(
        "-t",
        "--timeout",
        help="Timout between retries when trying to reach locally deployed endpoint",
        default=7,
    ),
    click.option(
        "-r",
        "--retries",
        help="Number of retries to reach locally deployed endpoint",
        default=7,
    ),
]


def shared_options_to_decorator(options, func):
    for option in reversed(options):
        func = option(func)
    return func


get_cmd_shared_options = functools.partial(shared_options_to_decorator, _shared_options)
filter_deploy_shared_options = functools.partial(
    shared_options_to_decorator, _filter_deploy_shared_options
)
filter_repo_shared_options = functools.partial(
    shared_options_to_decorator, _filter_repo_shared_options
)
validate_project_shared_options = functools.partial(
    shared_options_to_decorator, _validate_project_shared_options
)
validate_integration_shared_options = functools.partial(
    shared_options_to_decorator, _validate_integration_shared_options
)


def to_uuid(_, __, x):
    return UUID4(x) if x else None


def get_secrets(secrets_repository) -> Optional[Dict[str, str]]:
    if secrets_repository is not None:
        secrets_manager = SecretsManager(secrets_repository=secrets_repository)
        return secrets_manager.get_secrets()
    else:
        return None



import http
from typing import List

from tktl.core.exceptions.base import (
    TaktileCliError,
    TaktileRuntimeError,
    TaktileSdkError,
)


class TimeoutError(TaktileCliError):
    pass


class EndpointInstantiationError(TaktileRuntimeError):
    pass


class AuthenticationError(TaktileSdkError):
    pass


class UnlazifiedParquetError(TaktileRuntimeError):
    pass


class UnsupportedInputTypeException(TaktileSdkError):
    pass


class MissingDocker(TaktileCliError):
    pass


class ValidationException(TaktileRuntimeError):
    pass


class ProjectValidationException(TaktileCliError):
    pass


class HTTPException(Exception):  # TODO - Potential Cleanup
    def __init__(self, status_code: int, detail: str = None) -> None:
        if detail is None:
            detail = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.detail = detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"


class APISwitchException(TaktileSdkError):
    # TODO: Remove this in taktile-cli 1.0
    pass


class APIClientException(HTTPException):
    pass

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"


class APIClientExceptionRetryable(APIClientException):
    pass


class UserRepoValidationException(TaktileCliError):
    def __init__(
        self, missing_files: List, missing_directories: List, missing_config: bool
    ):
        self.missing_files = missing_files
        self.missing_directories = missing_directories
        self.missing_config = missing_config


class NoContentsFoundException(UserRepoValidationException):
    pass


def validate_config(fn):
    from tktl.core.managers.project import ProjectManager

    def wrapper(*args, **kwargs):
        if ProjectManager.get_config() is None:
            raise TaktileCliError(
                "No configuration found. Are you sure you have a tktl.yaml file in "
                "your current directory? Run tktl init to start a new project"
            )
        return fn(*args, **kwargs)

    return wrapper



import os
import time
from typing import Dict

import docker  # type: ignore
import requests
from requests.exceptions import RequestException

from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand
from tktl.core.config import settings
from tktl.core.exceptions import APIClientException, MissingDocker
from tktl.core.loggers import LOG, MUTE_LOG

TESTING_DOCKERFILE = "Dockerfile.taktile-cli-testing"
MULTI_STAGE_BUILD_STEP_NAME = "build_step"


class DockerManager:
    def __init__(self, path):
        try:
            self._client = docker.from_env()
            self._path = path
        except docker.errors.DockerException as err:
            raise MissingDocker from err

    def get_docker_file(self) -> str:
        with open(os.path.join(self._path, ".buildfile")) as fp:
            return fp.read()

    def stream_logs(self, container) -> None:
        for line in container.logs(stream=True):
            LOG.log(f"> {line.decode()}".strip())

    def remove_image(self, image):
        self._client.images.remove(image, force=True)

    def build_image(
        self, dockerfile: str, cache: bool, prune: bool, buildargs: Dict = None
    ) -> str:

        if prune:
            try:
                image = self._client.images.get("taktile-cli-test:latest")
                self._client.images.remove(image.id)
            except docker.errors.ImageNotFound:
                pass

        if not cache:
            LOG.log("Building without Docker cache. This may take some time...")

        LOG.log(
            f"Multistage dockerfile, building target {MULTI_STAGE_BUILD_STEP_NAME}..."
        )
        image, stream = self._client.images.build(
            path=self._path,
            dockerfile=dockerfile,
            tag="taktile-cli-test",
            nocache=not cache,
            target=MULTI_STAGE_BUILD_STEP_NAME,
            buildargs=buildargs,
            forcerm=True,
        )

        for chunk in stream:
            if "stream" in chunk:
                for line in chunk["stream"].splitlines():
                    LOG.trace("> " + line)

        return image.id

    def test_import(self, image_id: str):
        client_name = "tktl" if not settings.FUTURE_ENDPOINTS else "client"
        command = f"python -c 'from src.endpoints import {client_name}'"
        try:
            container = self._client.containers.run(image_id, command, detach=True,)
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def test_endpoint_sample(self, image_id: str):
        try:
            command = f"python /app/scripts/endpoint_sample_test{'_future' if settings.FUTURE_ENDPOINTS else ''}.py"
            container = self._client.containers.run(image_id, command, detach=True)
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def test_unittest(self, image_id: str):
        try:
            container = self._client.containers.run(
                image_id, "python -m pytest ./user_tests/", detach=True
            )
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def run_rest_container(
        self, image_id: str, detach: bool = True, auth_enabled: bool = True
    ):
        return self._client.containers.run(
            image_id,
            detach=detach,
            entrypoint="/start-rest.sh",
            environment={
                "AUTH_ENABLED": auth_enabled,
                "TAKTILE_GIT_SHA": "local-run",
                "TAKTILE_GIT_REF": "local-run",
            },
            ports={"80/tcp": 8080},
            remove=True,
            stderr=True,
            stdout=True,
        )

    def run_arrow_container(
        self, image_id: str, detach: bool = True, auth_enabled: bool = True
    ):
        return self._client.containers.run(
            image_id,
            detach=detach,
            entrypoint="/start-flight.sh",
            environment={
                "AUTH_ENABLED": auth_enabled,
                "TAKTILE_GIT_SHA": "local-run",
                "TAKTILE_GIT_REF": "local-run",
            },
            ports={"5005/tcp": 5005},
            remove=True,
            stderr=True,
            stdout=True,
        )

    def run_containers(
        self, image_id: str, detach: bool = True, auth_enabled: bool = True
    ):
        arrow_container = self.run_arrow_container(
            image_id=image_id, detach=detach, auth_enabled=auth_enabled
        )
        rest_container = self.run_rest_container(
            image_id=image_id, detach=detach, auth_enabled=auth_enabled
        )
        return arrow_container, rest_container

    def run_profiling_container(self):
        try:
            image_name = f"taktile/taktile-profiler:{_get_profiling_version()}"
            LOG.debug(f"Using Profiling Image {image_name}")
            container = self._client.containers.run(
                image_name,
                entrypoint="profiler profile -l",
                network_mode="host",
                detach=True,
            )
            self.stream_logs(container)
            status = container.wait()
            return status, container
        finally:
            container.remove()

    def run_and_check_health(
        self,
        image_id: str,
        kill_on_success: bool = False,
        auth_enabled: bool = True,
        timeout: int = 7,
        retries: int = 7,
    ):
        arrow_container, rest_container = self.run_containers(
            image_id=image_id, detach=True, auth_enabled=auth_enabled
        )
        grpc_health_cmd = GetGrpcHealthCommand(
            branch_name="",
            repository="",
            local=True,
            logger=MUTE_LOG if not LOG.VERBOSE else LOG,
            skip_auth=True,
        )
        rest_health_cmd = GetRestHealthCommand(
            branch_name="",
            repository="",
            local=True,
            logger=MUTE_LOG if not LOG.VERBOSE else LOG,
            skip_auth=True,
        )

        try:
            for _ in range(retries):
                try:
                    time.sleep(timeout)
                    rest_response = rest_health_cmd.execute()
                    grpc_response = grpc_health_cmd.execute()
                    return rest_response, grpc_response, arrow_container, rest_container
                except (APIClientException, Exception):
                    pass

            return None, None, arrow_container, rest_container
        finally:
            if kill_on_success:
                arrow_container.kill()
                rest_container.kill()


def _get_profiling_version() -> str:
    if settings.TAKTILE_PROFILING_VERSION:
        return settings.TAKTILE_PROFILING_VERSION
    else:
        try:
            res = requests.get(f"{settings.LOCAL_REST_ENDPOINT}/info").json()
            if isinstance(res, list):  # Not Future
                profiling_version = "latest"
            else:  # Future
                profiling_version = res.get("profiling", "latest")
        except RequestException as e:
            LOG.debug(f"Failed to fetch profiling version from Info {e}")
            profiling_version = "latest"
        return profiling_version
